#ifndef __CLASSIFIER_H
#define __CLASSIFIER_H

//Liste zum Gebrauch
#include "LinkedList.h"


#ifdef Arduino_h
#define _print(message) {Serial.print(message);}
#else
#define _print(message) {std::cout << message;}
#endif

//swap fkt
template <class T>
class Swapper
// to replicate std::swap, as it does not exist on Arduino
{
public:
	static void swap(T& a, T& b) {
		T c(a); a = b; b = c;
	}

private:
	Swapper() {}
};


template < typename element>
class Classifier
{
  private:
    struct Entry{
      //zB 5 elemente int
      element *x;
      // 1 label y size_t zB int
      size_t y;
    };

  size_t _numClasses;
  size_t _k, _entrysize, _memsize;
  //zeiger 1. eintrag
  Entry *_memory_x;
  bool _invalid;
  // 1. freier eintrag im speicher
  Entry *_vacant_x_mem_p;  // points to unused x memory

  inline double _calculateEucliedanDistance(const element *elem1, const element *elem2);
  void _allocateMemory();
  void _deallocateMemory();

  public:
  //konstruktor klassifizerer 
  Classifier(const size_t numClasses, const size_t k, const size_t memsize, const size_t entrysize,
        const size_t numEntries = 0, const element *x = NULL, const size_t *y = NULL);
  //
  Classifier(const Classifier& other);
  Classifier& operator=(Classifier rhs);
  ~Classifier(void);

  void _swap(Classifier<element>& first, Classifier<element>& second);

  //mehrere lernen also batch (numentries)
  int learn(const size_t numEntries, const element *x, const size_t *y);
  int learn(const element *x, const size_t y);
  //löschen
  inline void flush(void);
  //4....klasse
  inline void add_class(void);
  //gib mir label
  int classify(const element *data, size_t *const counterp = NULL);
  
  void print_memory(void);

  bool isInvalid();

};

// Public Methods //////////////////////////////////////////////////////////////
template < typename element >
Classifier<element>::Classifier(const size_t numClasses, const size_t k, const size_t memsize, const size_t entrysize,
  const size_t numEntries, const element *x, const size_t *y)
{
  /* catch parameter errors */
  if (numClasses < 1 || memsize < 1 || entrysize < 1) {
    _print("Classifier Constructor Exception: Invalid arguments. "
      "Must have numClasses, memsize and entrysize >= 1\r\n");
    _invalid = 1;
    return;
  }

  _invalid = 0;
  _numClasses = numClasses;
  _k = k;
  _entrysize = entrysize;
  _memsize = memsize;

  /* allocate memory for x */
  _allocateMemory();

  if (numEntries > 0)
    // if previous knowledge exists ...
    learn(numEntries, x, y);

#ifdef CLASSIFIER_DEBUG
  else 
    // if no previous knowledge exists ..
    _print("Empty call\r\n");
#endif
}

template < typename element>
Classifier<element>::Classifier(const Classifier<element>&C)
	:	_numClasses(C._numClasses),
		_k(C._k),
		_entrysize(C._entrysize),
		_memsize(C._memsize),
		_invalid(C._invalid)
{
	// allocate new memory
	_allocateMemory();

	_vacant_x_mem_p = _memory_x + (C._vacant_x_mem_p - C._memory_x);

	for (Entry* mp = C._memory_x; mp < C._vacant_x_mem_p; mp++) {
		memcpy(_memory_x[mp - C._memory_x].x, mp->x, _entrysize * sizeof(element));
		_memory_x[mp - C._memory_x].y = C._memory_x[mp - C._memory_x].y;
	}

}

template < typename element>
Classifier<element>& Classifier<element>::operator=(Classifier<element>C) {
	// see https://stackoverflow.com/questions/3279543/what-is-the-copy-and-swap-idiom

	_swap(*this, C);
	return *this;
}

template < typename element>
Classifier<element>::~Classifier(){
  _deallocateMemory();
}

template < typename element>
void Classifier<element>::_swap(Classifier<element>& first, Classifier<element>& second)
{
	Swapper<size_t>::swap(first._numClasses, second._numClasses);
	Swapper<size_t>::swap(first._k, second._k);
	Swapper<size_t>::swap(first._entrysize, second._entrysize);
	Swapper<size_t>::swap(first._memsize, second._memsize);
	Swapper<bool>::swap(first._invalid, second._invalid);
	Swapper<Classifier<element>::Entry *>::swap(first._memory_x, second._memory_x);
	Swapper<Classifier<element>::Entry*>::swap(first._vacant_x_mem_p, second._vacant_x_mem_p);

}

template < typename element>
int Classifier<element>::learn(const size_t numEntries, const element *x, const size_t *y) {

  if (isInvalid()) {
    return -1;
  }

  if (numEntries == 0) {
#ifdef CLASSIFIER_DEBUG
    _print("Empty call to learn. Nothing learned.\r\n");
#endif
    return 0;
  }

  if (numEntries && (x == NULL || y == NULL)) {
    _print("Classifier Learn Exception: Invalid arguments. "
      "Input pointer to entries is NULL.\r\n");
    return -1;
  }
  /* catch parameter error */
  for (size_t i = 0; i < numEntries; i++)
    if (y[i] >= _numClasses) {
      _print("Classifier Learn Exception: Invalid arguments. "
        "Numbers in y not in range [0,numClasses]\r\n"
        "Maybe memory overflow in y (numEntries to small)?\r\n");
      return -1;
    }

  // for every new entry that has to be inserted
  for (size_t i = 0; i < numEntries; i++) {
    // if there is still free space in memory left, insert
    if (_vacant_x_mem_p < (_memory_x + _memsize)) {

#ifdef CLASSIFIER_DEBUG
      _print("Learn: Inserted entry at place ");
      _print(_vacant_x_mem_p - _memory_x);
      _print(".\r\n");
#endif

      memcpy(_vacant_x_mem_p->x, x + (i * _entrysize), _entrysize * sizeof(element));
      _vacant_x_mem_p->y = y[i];
      _vacant_x_mem_p++;
    }
    else {
      /* 
         if there is no free space left, an old entry has to be deleted and replaced.
         We search for an old entry to be replaced by the following model:
         - We only replace an entry if it is classified correctly by KNN and with maximum confidence
         - The new entry also has to be classified correctly by KNN and with maximum confidence. Otherwise,
         it is skipped.

      */

      /* First, we classify the new entry to see whether it is classified correctly or not*/

      size_t *newc_details = new size_t[_numClasses];
      int newc_result = classify(x + (i * _entrysize), newc_details);   

        // if it is classified correctly and with maximum confidence
      if (newc_result == y[i] && newc_details[y[i]] == _k)
        // look at all entries in memory, one after another
        for (Entry *e = _memory_x; e < _vacant_x_mem_p; e++) {
          // classify them
          size_t *oldc_details = new size_t[_numClasses];
          int oldc_result = classify(e->x, oldc_details);

          // if the old entry is classified correctly and the confidence is maximal
          if (oldc_result == (int)e->y && oldc_details[oldc_result] == _k) {
            // replace old entry by new one
#ifdef CLASSIFIER_DEBUG
            _print("Learn: Replaced entry ");
            _print(i);
            _print(" ( class ");
            _print(e->y);
            _print(") by new one (class ");
            _print(y[i]);
            _print(").\r\n");
#endif

            memcpy(e->x, x + (i * _entrysize), _entrysize * sizeof(element));
            e->y = y[i];
            delete[] oldc_details;
            delete[] newc_details;
            goto continue_outer;
          }
          delete[] oldc_details;
        }
      delete[] newc_details;
#ifdef CLASSIFIER_DEBUG
      _print("Learn: Entry skipped.\r\n");
#endif      
    }
  continue_outer:;
  }
  return 0;
      



#ifdef CLASSIFIER_DEBUG
  _print("Success memory copy\r\n");
#endif
  return 0;
}

template < typename element>
int Classifier<element>::learn(const element *x, const size_t y) {
  return learn(1, x, &y);
}

template < typename element>
inline void Classifier<element>::flush() {
   _vacant_x_mem_p = _memory_x;
}

template < typename element>
inline void Classifier<element>::add_class() {
	_numClasses++;
}

template < typename element>
int Classifier<element>::classify(const element *data, size_t *const counterp) {
  /*
  Classifies the given data and returns the class number.
  If given a pointer "counterp" pointing to memory of size (numClasses * sizeof(size_t)),
  classify will store the detailed number of occurences of each class in that array
  */

  /*
    To save memory and execution time, this function calculates all eucliean
    distances between 'data' and the entries in memory, but saves only the k shortest.
    A pointer to each entry is stored together with its distance in a struct ewd.
    The k structs with shortest distance are stored in eList.
    In eList, the 'ewd's are stored with ascending distance. Whenever a new entry
    must be stored, it is inserted at the right place in the list.

    If 'data' points to an element in the memory, it is classified using all
    _other_ elements BUT that one.


  */

  struct ewd { // 'ewd': entry with distance
    Entry *e;
    double distance;
  };
  LinkedList<ewd*> eList;
  double curDistance;

  if (isInvalid()) {
    return -1;
  }

  // calculate euclidean distances and save k elements with shortest in list
  for (Entry *p = _memory_x; p < _vacant_x_mem_p; p++) {
    if (data == p->x)
      // If 'data' points to an element in the memory, it is classified using all
      // _other_ elements BUT that one.
      continue;
    curDistance = _calculateEucliedanDistance(p->x, data);
    if (eList.moveToStart()) {
      if (!(curDistance < eList.Last()->distance) && (eList.getLength() == (int)_k))
        // list is full and current entry's distance is larger than largest in list, so discard entry
        continue;

      // make a new entry, you'll need it.  
      ewd *newEntry = new ewd();
      newEntry->e = p;
      newEntry->distance = curDistance;

      // move down list until larger distance than the current entry ones is found.
      while ((eList.getCurrent()->distance) < curDistance && eList.next());

      if ((eList.getCurrent()->distance) >= curDistance) 
        // eList currently points to the first ewd whose distance is equal or larger than the current one.
        // Therefore, create a new ewd and insert it in list before eList pointer.
        eList.InsertBefore(newEntry);
    
      else
        // eList currently points to the largest entry of the list, but that is still smaller than the current entry's
        eList.Append(newEntry);
    }
    else
    {
      // List was empty, so store first entry.
      ewd *newEntry = new ewd();
      newEntry->e = p;
      newEntry->distance = curDistance;
      eList.Append(newEntry);
    }

    // When k entries are already present in list before insertion, the list can become too large.
    // Therefore, the entry with largest distance is now removed.
    if (eList.getLength() > (int)_k) {
      delete eList.Last();
      eList.DeleteLast();
    }
  }

  // for elements in list, retrieve their class. determine majority
  size_t *counter = new size_t[_numClasses];
  for (size_t i = 0; i < _numClasses; i++)
    counter[i] = 0;

  if (eList.moveToStart())
  do {
    ewd *current = eList.getCurrent();
    counter[current->e->y]++;
    delete current;
  } while (eList.next());

  int majorityClass = 0;
  for (size_t i = 1; i < _numClasses; i++) {
    if (counter[i] > counter[majorityClass])
      majorityClass = i;
  }

  if (counterp)
	memcpy(counterp, counter, sizeof(size_t) * _numClasses);
  delete[] counter;
  return majorityClass;
  return 0;
}

template<typename element>
void Classifier<element>::print_memory() {

  if (isInvalid()) {
    return;
  }

  _print("Memory dump X:\r\nClass\tContent\r\n");

  for (Entry *p = _memory_x; p < _vacant_x_mem_p; p++) {
    _print(p->y); _print("\t");
    for (element *q = p->x; q < (p->x + _entrysize); q++) {
      _print(*q); _print("\t");
    }
    _print("\r\n");
  }
  size_t freeSpace = (_memory_x + _memsize) - _vacant_x_mem_p;

  if (freeSpace) {
    _print("And "); _print(freeSpace); _print(" unused entry block");
    if (freeSpace > 1)
      _print("s");
    _print(".\r\n");
  }


  _print("\r\nMemory dump sorted by class:\r\n");

  for (size_t i = 0; i < _numClasses; i++) {
    _print("Class "); _print(i); _print("\r\n");

    for (Entry *p = _memory_x; p < _vacant_x_mem_p; p++) 
      if (p->y == i) {
        for (element *x = p->x; x < (p->x + _entrysize); x++) {
          _print(*x); _print("\t");
        }
        _print("\r\n");
      }
    _print("\r\n");
    
  }
}

template<typename element>
bool Classifier<element>::isInvalid() {
  if (_invalid) {
    _print("Due to wrong initialization on your side, this object is invalid and useless.\r\n"
      "Please delete, check your constructor call and retry.");
  }
  return _invalid;
}

// Private Methods //////////////////////////////////////////////////////////////
template<typename element>
inline double Classifier<element>::_calculateEucliedanDistance(const element * elem1, const element * elem2){
  double *squared_difference = new double[_entrysize];
  size_t i;
  double result = 0;

  for (i = 0; i < _entrysize; i++)
    squared_difference[i] = pow(elem1[i] - elem2[i], 2);
  for (i = 0; i < _entrysize; i++)
    result += squared_difference[i];
  result = sqrt(result);

  delete[] squared_difference;
  return result;

}

template<typename element>
void Classifier<element>::_allocateMemory() {
  _memory_x = new Entry[_memsize];

  for (unsigned i = 0; i < _memsize; i++)
    _memory_x[i].x = new element[_entrysize];

  _vacant_x_mem_p = _memory_x;
}

template<typename element>
void Classifier<element>::_deallocateMemory() {
  for (unsigned i = 0; i < _memsize; i++)
    delete[] _memory_x[i].x;

  delete[] _memory_x;
}

#endif
